'''
Function:
    定义Sequential模型
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from .module import Module


'''定义Sequential模型'''
class Sequential(Module):
    def __init__(self, module_list=[], **kwargs):
        super(Sequential, self).__init__(**kwargs)
        self.module_dict = {}
        for idx, module in enumerate(module_list):
            self.module_dict.update({str(idx): module})
    '''添加模型'''
    def addmodule(self, name, module):
        self.module_dict.update({name: module})
    '''前向传播'''
    def forward(self, x):
        for module in self.module_dict.values():
            x = module(x)
        return x
    '''反向传播'''
    def backward(self, accumulated_gradient):
        for module in reversed(list(self.module_dict.values())):
            accumulated_gradient = module.backward(accumulated_gradient)
        return accumulated_gradient
    '''返回参数数量'''
    def parameters(self):
        num_paramters = 0
        for module in self.module_dict.values(): 
            num_paramters += module.parameters()
        return num_paramters
    '''返回所有modules'''
    def modules(self):
        return self.module_dict


'''
Function:
    定义一些常见的评估函数
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np


'''均方误差损失'''
class MSELoss():
    def __init__(self, reduction='mean'):
        assert reduction in ['sum', 'none', 'mean']
        self.reduction = reduction
        self.storage = {}
    '''定义前向传播'''
    def __call__(self, preditions, targets):
        self.storage.update({
            'preditions': preditions, 'targets': targets
        })
        loss = 0.5 * (preditions - targets) ** 2
        if self.reduction == 'none': return loss
        loss = getattr(loss, self.reduction)()
        return loss
    '''定义反向传播'''
    def backward(self):
        gradient = self.storage['preditions'] - self.storage['targets']
        if self.reduction == 'mean': gradient /= gradient.shape[0]
        return gradient


'''交叉熵损失'''
class CrossEntropy():
    def __init__(self, reduction='mean', eps=1e-12):
        assert reduction in ['sum', 'none', 'mean']
        self.reduction = reduction
        self.eps = eps
        self.storage = {}
    '''定义前向传播'''
    def __call__(self, preditions, targets):
        self.storage.update({
            'preditions': preditions, 'targets': targets
        })
        preditions = np.clip(preditions, self.eps, 1 - self.eps)
        loss = - targets * np.log(preditions) - (1 - targets) * np.log(1 - preditions)
        if self.reduction == 'none': return loss
        loss = getattr(loss, self.reduction)()
        return loss
    '''定义反向传播'''
    def backward(self):
        preditions, targets = self.storage['preditions'], self.storage['targets']
        preditions = np.clip(preditions, self.eps, 1 - self.eps)
        gradient = - (targets / preditions) + (1 - targets) / (1 - preditions)
        if self.reduction == 'mean': gradient /= gradient.shape[0]
        return gradient


'''initialize'''
from .linear import (
    Linear
)
from .module import (
    Module
)
from .convolution import (
    Conv2d
)
from .sequential import (
    Sequential
)
from .criterion import (
    MSELoss, CrossEntropy
)
from .activation import (
    Softmax
)


'''
Function:
    网络模型基础类
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np


'''网络模型基础类'''
class Module():
    def __init__(self, **kwargs):
        for key, value in kwargs.items(): setattr(self, key, value)
        # 默认为可训练状态
        self.training = True
        # 保存一些必要的数据用于反向传播
        self.storage = {}
    '''前向传播调用'''
    def __call__(self, x, **kwargs):
        self.storage.update({'x': x})
        x = self.forward(x, **kwargs)
        return x
    '''定义前向传播'''
    def forward(self):
        raise NotImplementedError('not to be implemented')
    '''定义反向传播'''
    def backward(self):
        raise NotImplementedError('not to be implemented')
    '''返回参数数量'''
    def parameters(self):
        return 0
    '''返回模型的名字'''
    def name(self):
        return self.__class__.__name__
    '''设置为训练状态'''
    def train(self, mode=True):
        def apply(module_dict, mode):
            for module in module_dict.values():
                if isinstance(module, dict):
                    apply(module, mode)
                else:
                    setattr(module, 'training', mode)
        module_dict = self.modules()
        apply(module_dict, mode)
    '''设置为测试状态'''
    def eval(self):
        self.train(mode=False)
    '''根据使用的优化器更新参数'''
    def update(self):
        raise NotImplementedError('not to be implemented')
    '''返回所有modules'''
    def modules(self):
        module_dict, attrs = {self.name(): self}, dir(self)
        for attr in attrs:
            value = getattr(self, attr)
            if isinstance(value, Module):
                module_dict[f'{self.name()}.{attr}'] = value.modules()
        return module_dict


'''
Function:
    定义SGD优化器
Author:
    Zhenchao
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from .baseoptimizer import BaseOptimizer


'''定义SGD优化器'''
class SGD(BaseOptimizer):
    def __init__(self, structure, criterion, learning_rate=0.01, momentum=0):
        super(SGD, self).__init__(
            structure=structure, criterion=criterion, learning_rate=learning_rate, momentum=momentum
        )
        # 所有网络层都使用优化器的update函数
        self.applyupdate(self.structure.modules())
    '''更新函数'''
    def update(self, params, grads, direction):
        direction = self.momentum * direction + (1 - self.momentum) * grads
        params = params - self.learning_rate * direction
        return {
            'params': params, 'direction': direction,
        }


'''initialize'''
from . import nn
from . import optim
from . import functional


'''package相关变量'''
__title__ = 'pytoydl'
__description__ = 'Pytoydl: A toy deep learning framework built upon numpy.'
__url__ = 'https://github.com/CharlesPikachu/pytoydl'
__version__ = '0.0.1'
__author__ = 'Zhenchao Jin'
__email__ = 'charlesblwx@gmail.com'
__license__ = 'Apache License 2.0'
__copyright__ = 'Copyright 2022 Zhenchao Jin'


'''
Function:
    定义激活函数
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from .module import Module


'''softmax'''
class Softmax(Module):
    def __init__(self, dim=-1):
        super(Softmax, self).__init__(dim=dim)
    '''定义前向传播'''
    def forward(self, x):
        ex = np.exp(x - np.max(x, axis=self.dim, keepdims=True))
        return ex / np.sum(ex, axis=-1, keepdims=True)
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        gradient = self.forward(self.storage['x'])
        gradient = gradient * (1 - gradient)
        return accumulated_gradient * gradient


'''
Function:
    实现全连接层
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import math
import numpy as np
from .module import Module


'''全连接层'''
class Linear(Module):
    def __init__(self, in_features, out_features, bias=True):
        super(Linear, self).__init__(
            in_features=in_features, out_features=out_features, bias=True
        )
        # 初始化权重
        self.weight = np.random.uniform(
            - 1 / math.sqrt(in_features), 1 / math.sqrt(in_features), (in_features, out_features)
        )
        self.bias = np.zeros((1, out_features))
        # 初始化storage
        self.storage.update({
            'direction': {
                'weight': np.zeros(np.shape(self.weight)), 
                'bias': np.zeros(np.shape(self.bias))
            }
        })
    '''定义前向传播'''
    def forward(self, x):
        feats = x.dot(self.weight) + self.bias
        return feats
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        weight = self.weight
        if self.training:
            # 计算梯度
            grad_w = self.storage['x'].T.dot(accumulated_gradient)
            grad_b = np.sum(accumulated_gradient, axis=0, keepdims=True)
            # 根据梯度更新weight
            results = self.update(self.weight, grad_w, self.storage['direction']['weight'])
            self.weight, self.storage['direction']['weight'] = results['params'], results['direction']
            # 根据梯度更新bias
            results = self.update(self.bias, grad_b, self.storage['direction']['bias'])
            self.bias, self.storage['direction']['bias'] = results['params'], results['direction']
        return accumulated_gradient.dot(weight.T)
    '''返回参数数量'''
    def parameters(self):
        return np.prod(self.weight.shape) + np.prod(self.bias.shape)


'''
Function:
    定义基类优化器
Author:
    Zhenchao
微信公众号:
    Charles的皮卡丘
'''
import numpy as np


'''定义基类优化器'''
class BaseOptimizer():
    def __init__(self, **kwargs):
        for key, value in kwargs.items(): setattr(self, key, value)
    '''所有网络层都使用优化器的update函数'''
    def applyupdate(self, module_dict):
        for module in module_dict.values():
            if isinstance(module, dict):
                self.applyupdate(module)
            else:
                setattr(module, 'update', self.update)
    '''梯度更新函数'''
    def update(self, params, grads, direction):
        raise NotImplementedError('not to be implemented')
    '''参数更新'''
    def step(self):
        self.structure.backward(self.criterion.backward())


'''
Function:
    实现卷积层
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from .module import Module


'''2d卷积'''
class Conv2d(Module):
    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=True):
        super(Conv2d, self).__init__(
            in_channels=int(in_channels), out_channels=int(out_channels), kernel_size=int(kernel_size),
            stride=int(stride), padding=int(padding), bias=int(bias)
        )
        assert (self.stride > 0) and (self.padding >=0) and (self.kernel_size >= 1) and (self.in_channels > 0) and (self.out_channels > 0)
